@@ -15,6 +15,10 @@
 # limitations under the License.
 #
 
+
+
+
+
 """
 In-memory persistent stub for the Python datastore API. Gets, queries,
 and searches are implemented as in-memory scans over all entities.
@@ -35,6 +39,12 @@ per-transaction, so they should only be used by one tx at a time.
 
 
 
+
+
+
+
+
+
 import datetime
 import logging
 import os
@@ -43,6 +53,8 @@ import sys
 import tempfile
 import threading
 
+
+
 import cPickle as pickle
 
 from google.appengine.api import api_base_pb
@@ -61,17 +73,24 @@ from google.net.proto import ProtocolBuffer
 from google.appengine.datastore import entity_pb
 
 
+
 entity_pb.Reference.__hash__ = lambda self: hash(self.Encode())
 datastore_pb.Query.__hash__ = lambda self: hash(self.Encode())
 datastore_pb.Transaction.__hash__ = lambda self: hash(self.Encode())
 
 
+
+
 _MAX_QUERY_COMPONENTS = 100
 
 
+
+
 _BATCH_SIZE = 20
 
 
+
+
 _MAX_ACTIONS_PER_TXN = 5
 
 
@@ -92,8 +111,14 @@ class _StoredEntity(object):
     """
     self.protobuf = entity
 
+
+
+
+
+
     self.encoded_protobuf = entity.Encode()
 
+
     self.native = datastore.Entity._FromPb(entity,
                                            validate_reserved_properties=False)
 
@@ -128,6 +153,7 @@ class KindPseudoKind(object):
         query.app(), query.name_space())
     kinds = []
 
+
     for app_namespace, kind in entities:
       if app_namespace != app_namespace_str: continue
       if not kind_range.Contains(kind): continue
@@ -187,6 +213,8 @@ class PropertyPseudoKind(object):
     for app_namespace, kind in entities:
       if app_namespace != app_namespace_str: continue
 
+
+
       (start_cmp, end_cmp) = property_range.MapExtremes(
           lambda extreme, inclusive, is_end: cmp(kind, extreme[0]))
       if not((start_cmp is None or start_cmp >= 0) and
@@ -195,20 +223,26 @@ class PropertyPseudoKind(object):
 
       app_kind = (app_namespace_str, kind)
 
+
       kind_properties = self.filestub._GetSchemaCache(app_kind, usekey)
       if not kind_properties:
         kind_properties = []
         kind_key = datastore_types.Key.from_path(KindPseudoKind.name, kind)
+
         props = {}
 
+
+
         for entity in entities[app_kind].values():
           for prop in entity.protobuf.property_list():
             prop_name = prop.name()
+
             if (prop_name in
                 datastore_stub_util.GetInvisibleSpecialPropertyNames()):
               continue
             if prop_name not in props:
               props[prop_name] = set()
+
             native_value = entity.native[prop_name]
             if not isinstance(native_value, list):
               native_value = [native_value]
@@ -219,6 +253,7 @@ class PropertyPseudoKind(object):
               else:
                 logging.warning('Unexpected value of class %s in datastore', value.__class__)
 
+
         for prop in sorted(props):
           property_e = datastore.Entity(self.name, name=prop, parent=kind_key)
           kind_properties.append(property_e)
@@ -230,6 +265,7 @@ class PropertyPseudoKind(object):
 
         self.filestub._SetSchemaCache(app_kind, usekey, kind_properties)
 
+
       def InQuery(property_e):
         return property_range.Contains((kind, property_e.key().name()))
       properties += filter(InQuery, kind_properties)
@@ -273,6 +309,7 @@ class NamespacePseudoKind(object):
       if app_id == app_str and namespace_range.Contains(namespace):
         namespaces.add(namespace)
 
+
     namespace_entities = []
     for namespace in namespaces:
       if namespace:
@@ -293,6 +330,9 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
   and is backed by files on disk.
   """
 
+
+
+
   _PROPERTY_TYPE_TAGS = {
     datastore_types.Blob: entity_pb.PropertyValue.kstringValue,
     bool: entity_pb.PropertyValue.kbooleanValue,
@@ -318,6 +358,13 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     users.User: entity_pb.PropertyValue.kUserValueGroup,
     }
 
+
+
+
+
+
+
+
   WRITE_ONLY = entity_pb.CompositeIndex.WRITE_ONLY
   READ_WRITE = entity_pb.CompositeIndex.READ_WRITE
   DELETED = entity_pb.CompositeIndex.DELETED
@@ -355,26 +402,55 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     super(DatastoreFileStub, self).__init__(service_name)
 
 
+
+
+
     assert isinstance(app_id, basestring) and app_id != ''
     self.__app_id = app_id
     self.__datastore_file = datastore_file
     self.SetTrusted(trusted)
 
+
+
+
+
+
+
+
     self.__entities = {}
 
+
+
+
     self.__schema_cache = {}
 
+
+
+
+
+
+
     self.__tx_snapshot = {}
 
+
+
     self.__tx_actions = []
 
+
+
+
     self.__queries = {}
 
+
     self.__transactions = set()
 
+
+
+
     self.__indexes = {}
     self.__require_indexes = require_indexes
 
+
     self.__query_history = {}
 
     self.__next_id = 1
@@ -388,6 +464,7 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     self.__file_lock = threading.Lock()
     self.__indexes_lock = threading.Lock()
 
+
     self.__pseudo_kinds = {}
     self._RegisterPseudoKind(KindPseudoKind())
     self._RegisterPseudoKind(PropertyPseudoKind(self))
@@ -427,6 +504,7 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     Raises:
       datastore_errors.BadRequestError: if this is not the stub for app_id.
     """
+
     assert app_id
     if not self.__trusted and app_id != self.__app_id:
       raise datastore_errors.BadRequestError(
@@ -441,6 +519,7 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     Raises:
       datastore_errors.BadRequestError: if the key is invalid
     """
+
     assert isinstance(key, entity_pb.Reference)
 
     self.__ValidateAppId(key.app())
@@ -493,6 +572,7 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
       self.__entities[app_kind] = {}
     self.__entities[app_kind][key] = _StoredEntity(entity)
 
+
     if app_kind in self.__schema_cache:
       del self.__schema_cache[app_kind]
 
@@ -530,6 +610,8 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
         except struct.error, e:
           if (sys.version_info[0:3] == (2, 5, 0)
               and e.message.startswith('unpack requires a string argument')):
+
+
             raise datastore_errors.InternalError(self.READ_PY250_MSG +
                                                  self.READ_ERROR_MSG %
                                                  (self.__datastore_file, e))
@@ -573,6 +655,8 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
           logging.warning('Could not read datastore data from %s', filename)
       except (AttributeError, LookupError, ImportError, NameError, TypeError,
               ValueError, struct.error, pickle.PickleError), e:
+
+
         raise datastore_errors.InternalError(
           'Could not read data from %s. Try running with the '
           '--clear_datastore flag. Cause:\n%r' % (filename, e))
@@ -587,8 +671,14 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     if not filename or filename == '/dev/null' or not obj:
       return
 
+
     descriptor, tmp_filename = tempfile.mkstemp(dir=os.path.dirname(filename))
     tmpfile = os.fdopen(descriptor, 'wb')
+
+
+
+
+
     pickler = pickle.Pickler(tmpfile, protocol=1)
     pickler.fast = True
     pickler.dump(obj)
@@ -598,8 +688,10 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     self.__file_lock.acquire()
     try:
       try:
+
         os.rename(tmp_filename, filename)
       except OSError:
+
         try:
           os.remove(filename)
         except:
@@ -622,11 +714,13 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     """Raises an exception if the given PB is not initialized and valid."""
     explanation = []
     assert pb.IsInitialized(explanation), explanation
+
     pb.Encode()
 
   def QueryHistory(self):
     """Returns a dict that maps Query PBs to times they've been run.
     """
+
     return dict((pb, times) for pb, times in self.__query_history.items()
                 if pb.app() == self.__app_id)
 
@@ -645,13 +739,16 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     if put_request.has_transaction():
       self.__ValidateTransaction(put_request.transaction())
 
+
     clones = []
     for entity in put_request.entity_list():
+
       self.__ValidateKey(entity.key())
 
       clone = entity_pb.EntityProto()
       clone.CopyFrom(entity)
 
+
       for property in clone.property_list() + clone.raw_property_list():
         datastore_stub_util.FillUser(property)
 
@@ -662,11 +759,13 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
 
       last_path = clone.key().path().element_list()[-1]
       if last_path.id() == 0 and not last_path.has_name():
+
         self.__id_lock.acquire()
         last_path.set_id(self.__next_id)
         self.__next_id += 1
         self.__id_lock.release()
 
+
         assert clone.entity_group().element_size() == 0
         group = clone.mutable_entity_group()
         root = clone.key().path().element(0)
@@ -676,8 +775,11 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
         assert (clone.has_entity_group() and
                 clone.entity_group().element_size() > 0)
 
+
+
       datastore_stub_util.PrepareSpecialPropertiesForStore(clone)
 
+
     self.__entities_lock.acquire()
 
     try:
@@ -686,17 +788,21 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     finally:
       self.__entities_lock.release()
 
+
     if not put_request.has_transaction():
       self.__WriteDatastore()
 
     put_response.key_list().extend([c.key() for c in clones])
 
+
   def _Dynamic_Touch(self, get_request, get_response):
     pass
 
   def _Dynamic_Get(self, get_request, get_response):
     if get_request.has_transaction():
       self.__ValidateTransaction(get_request.transaction())
+
+
       entities = self.__tx_snapshot
     else:
       entities = self.__entities
@@ -707,8 +813,12 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
 
       group = get_response.add_entity()
       try:
+
+
+
         entity = entities[app_kind][key].protobuf
       except KeyError:
+
         entity = None
 
       if entity:
@@ -729,12 +839,16 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
         try:
           del self.__entities[app_kind][key]
           if not self.__entities[app_kind]:
+
             del self.__entities[app_kind]
 
+
           del self.__schema_cache[app_kind]
         except KeyError:
+
           pass
 
+
         if not delete_request.has_transaction():
           self.__WriteDatastore()
     finally:
@@ -744,6 +858,8 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
   def _Dynamic_RunQuery(self, query, query_result):
     if query.has_transaction():
       self.__ValidateTransaction(query.transaction())
+
+
       entities = self.__tx_snapshot
     else:
       entities = self.__entities
@@ -752,6 +868,7 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     namespace = query.name_space()
     self.__ValidateAppId(app_id)
 
+
     (filters, orders) = datastore_index.Normalize(query.filter_list(),
                                                   query.order_list())
     datastore_stub_util.ValidateQuery(query, filters, orders,
@@ -763,6 +880,7 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
       pseudo_kind = self.__pseudo_kinds[query.kind()]
 
     if not pseudo_kind and self.__require_indexes:
+
       required, kind, ancestor, props, num_eq_filters = datastore_index.CompositeIndexForQuery(query)
       if required:
         required_key = kind, ancestor, props
@@ -780,6 +898,7 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
           if required_key == index_key:
             break
           if num_eq_filters > 1 and (kind, ancestor) == index_key[:2]:
+
             this_props = index_key[2]
             this_eq_filters_set = set(this_props[:num_eq_filters])
             this_remaining_filters = this_props[num_eq_filters:]
@@ -787,16 +906,21 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
                 remaining_filters == this_remaining_filters):
               break
         else:
+
           raise apiproxy_errors.ApplicationError(
               datastore_pb.Error.NEED_INDEX,
               "This query requires a composite index that is not defined. "
               "You must update the index.yaml file in your application root.")
 
+
+
+
     try:
       query.set_app(app_id)
       datastore_types.SetNamespace(query, namespace)
       encoded = datastore_types.EncodeAppIdNamespace(app_id, namespace)
       if pseudo_kind:
+
         (results, filters, orders) = pseudo_kind.Query(entities, query,
                                                        filters, orders)
       elif query.has_kind():
@@ -810,6 +934,9 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     except KeyError:
       results = []
 
+
+
+
     if query.has_ancestor():
       ancestor_path = query.ancestor().path().element_list()
       def is_descendant(entity):
@@ -817,6 +944,9 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
         return path[:len(ancestor_path)] == ancestor_path
       results = filter(is_descendant, results)
 
+
+
+
     operators = {datastore_pb.Query_Filter.LESS_THAN:             '<',
                  datastore_pb.Query_Filter.LESS_THAN_OR_EQUAL:    '<=',
                  datastore_pb.Query_Filter.GREATER_THAN:          '>',
@@ -836,19 +966,28 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
         values = [values]
 
       for value in values:
+
+
         if type(value) not in datastore_types._RAW_PROPERTY_TYPES:
           return True
       return False
 
     for filt in filters:
+
       assert filt.op() != datastore_pb.Query_Filter.IN
 
+
       prop = filt.property(0).name().decode('utf-8')
       op = operators[filt.op()]
 
+
+
+
+
       filter_val_list = [datastore_types.FromPropertyPb(filter_prop)
                          for filter_prop in filt.property_list()]
 
+
       def passes_filter(entity):
         """Returns True if the entity passes the filter, False otherwise.
 
@@ -866,15 +1005,31 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
         if not isinstance(entity_vals, list):
           entity_vals = [entity_vals]
 
+
         for fixed_entity_val in entity_vals:
           for filter_val in filter_val_list:
+
+
+
+
+
+
+
+
+
             fixed_entity_type = self._PROPERTY_TYPE_TAGS.get(
               fixed_entity_val.__class__)
             filter_type = self._PROPERTY_TYPE_TAGS.get(filter_val.__class__)
             if fixed_entity_type == filter_type:
+
               comp = u'%r %s %r' % (fixed_entity_val, op, filter_val)
             elif op != '==':
+
+
               comp = '%r %s %r' % (fixed_entity_type, op, filter_type)
+
+
+
             else:
               continue
 
@@ -890,28 +1045,42 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
 
         return False
 
+
       results = filter(passes_filter, results)
 
+
+
+
     for order in orders:
       prop = order.property().decode('utf-8')
       results = [entity for entity in results if has_prop_indexed(entity, prop)]
 
+
+
+
     def order_compare_entities(a, b):
       """ Return a negative, zero or positive number depending on whether
       entity a is considered smaller than, equal to, or larger than b,
       according to the query's orderings. """
+
       cmped = 0
       for o in orders:
         prop = o.property().decode('utf-8')
 
+
+
+
+
         reverse = (o.direction() is datastore_pb.Query_Order.DESCENDING)
 
         a_val = datastore._GetPropertyValue(a, prop)
         if isinstance(a_val, list):
+
           a_val = sorted(a_val, order_compare_properties, reverse=reverse)[0]
 
         b_val = datastore._GetPropertyValue(b, prop)
         if isinstance(b_val, list):
+
           b_val = sorted(b_val, order_compare_properties, reverse=reverse)[0]
 
         cmped = order_compare_properties(a_val, b_val)
@@ -940,6 +1109,8 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
       on the type ordering used in the real datastore, which is based on the
       tag numbers in the PropertyValue PB.
       """
+
+
       if isinstance(x, datetime.datetime):
         x = datastore_types.DatetimeToTimestamp(x)
       if isinstance(y, datetime.datetime):
@@ -949,15 +1120,18 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
       y_type = self._PROPERTY_TYPE_TAGS.get(y.__class__)
 
       if x_type == y_type:
+
         try:
           return cmp(x, y)
         except TypeError:
           return 0
       else:
+
         return cmp(x_type, y_type)
 
     results.sort(order_compare_entities)
 
+
     clone = datastore_pb.Query()
     clone.CopyFrom(query)
     clone.clear_hint()
@@ -968,12 +1142,16 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     else:
       self.__query_history[clone] = 1
 
+
+
+
+
+
     results = [r._ToPb() for r in results]
     for result in results:
       datastore_stub_util.PrepareSpecialPropertiesForLoad(result)
     cursor = datastore_stub_util.ListCursor(query, results,
                                             order_compare_entities_pb)
-    self.__queries[cursor.cursor] = cursor
 
     if query.has_count():
       count = query.count()
@@ -986,21 +1164,28 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
                                query.offset(), compile=query.compile())
 
     if query.compile():
+
+
       compiled_query = query_result.mutable_compiled_query()
       compiled_query.set_keys_only(query.keys_only())
       compiled_query.mutable_primaryscan().set_index_name(query.Encode())
 
+    if query_result.has_cursor():
+      self.__queries[query_result.cursor().cursor()] = cursor
+
   def _Dynamic_Next(self, next_request, query_result):
     self.__ValidateAppId(next_request.cursor().app())
 
     cursor_handle = next_request.cursor().cursor()
 
+
     try:
       cursor = self.__queries[cursor_handle]
     except KeyError:
       raise apiproxy_errors.ApplicationError(
           datastore_pb.Error.BAD_REQUEST, 'Cursor %d not found' % cursor_handle)
 
+
     assert cursor.app == next_request.cursor().app()
 
     count = _BATCH_SIZE
@@ -1009,9 +1194,18 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     cursor.PopulateQueryResult(query_result,
                                count, next_request.offset(),
                                next_request.compile())
+    if not query_result.has_cursor():
+      del self.__queries[cursor_handle]
 
   def _Dynamic_Count(self, query, integer64proto):
     query_result = datastore_pb.QueryResult()
+
+
+
+
+
+
+
     self._Dynamic_RunQuery(query, query_result)
     cursor = query_result.cursor().cursor()
     integer64proto.set_value(min(self.__queries[cursor].Count(),
@@ -1046,6 +1240,8 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     """
 
 
+
+
     if ((len(self.__tx_actions) + request.add_request_size()) >
         _MAX_ACTIONS_PER_TXN):
       raise apiproxy_errors.ApplicationError(
@@ -1054,6 +1250,9 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
 
     new_actions = []
     for add_request in request.add_request_list():
+
+
+
       self.__ValidateTransaction(add_request.transaction())
       clone = taskqueue_service_pb.TaskQueueAddRequest()
       clone.CopyFrom(add_request)
@@ -1066,9 +1265,11 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     self.__ValidateTransaction(transaction)
 
     self.__tx_snapshot = {}
+
     try:
       self.__WriteDatastore()
 
+
       for action in self.__tx_actions:
         try:
           apiproxy_stub_map.MakeSyncCall(
@@ -1097,6 +1298,7 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     namespace_str = req.name_space()
     app_namespace_str = datastore_types.EncodeAppIdNamespace(app_str,
                                                              namespace_str)
+
     kinds = []
 
     for app_namespace, kind in self.__entities:
@@ -1111,13 +1313,17 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
         kinds.append(kind_pb)
         continue
 
+
       kind_pb = entity_pb.EntityProto()
       kind_pb.mutable_key().set_app('')
       kind_pb.mutable_key().mutable_path().add_element().set_type(kind)
       kind_pb.mutable_entity_group()
 
+
       props = {}
 
+
+
       for entity in self.__entities[app_kind].values():
         for prop in entity.protobuf.property_list():
           if (prop.name() in
@@ -1125,8 +1331,11 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
             continue
           if prop.name() not in props:
             props[prop.name()] = entity_pb.PropertyValue()
+
           props[prop.name()].MergeFrom(prop.value())
 
+
+
       for value_pb in props.values():
         if value_pb.has_int64value():
           value_pb.set_int64value(0)
@@ -1152,15 +1361,18 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
           pathelem.set_type('none')
           pathelem.set_name('none')
 
+
       for name, value_pb in props.items():
         prop_pb = kind_pb.add_property()
         prop_pb.set_name(name)
         prop_pb.set_multiple(False)
         prop_pb.mutable_value().CopyFrom(value_pb)
 
+
       kinds.append(kind_pb)
       self._SetSchemaCache(app_kind, "GetSchema", kind_pb)
 
+
     for kind_pb in kinds:
       kind = schema.add_kind()
       kind.CopyFrom(kind_pb)
@@ -1177,6 +1389,7 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
     if allocate_ids_request.has_size() and allocate_ids_request.has_max():
       raise apiproxy_errors.ApplicationError(datastore_pb.Error.BAD_REQUEST,
                                              'Both size and max cannot be set.')
+
     try:
       self.__id_lock.acquire()
       start = self.__next_id
@@ -1200,17 +1413,20 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
       raise apiproxy_errors.ApplicationError(datastore_pb.Error.BAD_REQUEST,
                                              'Index already exists.')
 
+
     self.__index_id_lock.acquire()
     index.set_id(self.__next_index_id)
     id_response.set_value(self.__next_index_id)
     self.__next_index_id += 1
     self.__index_id_lock.release()
 
+
     clone = entity_pb.CompositeIndex()
     clone.CopyFrom(index)
     app = index.app_id()
     clone.set_app_id(app)
 
+
     self.__indexes_lock.acquire()
     try:
       if app not in self.__indexes:
@@ -1239,6 +1455,7 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
           (entity_pb.CompositeIndex.State_Name(stored_index.state()),
           (entity_pb.CompositeIndex.State_Name(index.state()))))
 
+
     self.__indexes_lock.acquire()
     try:
       stored_index.set_state(index.state())
@@ -1252,6 +1469,7 @@ class DatastoreFileStub(apiproxy_stub.APIProxyStub):
       raise apiproxy_errors.ApplicationError(datastore_pb.Error.BAD_REQUEST,
                                              "Index doesn't exist.")
 
+
     app = index.app_id()
     self.__indexes_lock.acquire()
     try:
